import 'dart:async';
import 'dart:convert' as convert;

import 'package:apex_flutter_plugin/api/base_request.dart';
import 'package:apex_flutter_plugin/api/cipher/crypto.dart';
import 'package:apex_flutter_plugin/api/config.dart';
import 'package:apex_flutter_plugin/api/request.dart';
import 'package:apex_flutter_plugin/api/response.dart';
import 'package:apex_flutter_plugin/api/result.dart';
import 'package:connectivity/connectivity.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:http/http.dart' as http;

import 'cancellation_token.dart';
import 'retry_helper.dart';

class Api {
  ApiConfig config = ApiConfig("https://google.com/");
  http.Client client = http.Client();

  static final Api _singleton = Api._();

  factory Api() => _singleton;

  Api._();

  static Api get shared => _singleton;

  String _prepareRequestMessage(Crypto crypto, Request request) {
    _print("default parameters : " + config.defaultParameters.toString());

    if (config.defaultParameters != null) {
      request.addParams(config.defaultParameters);
    }

    String requestMessage = request.toJson();

    _print("raw request : " + requestMessage);

    if (config.encrypt) {
      requestMessage = crypto.encrypt(requestMessage);
      _print("encrypted request : " + requestMessage);
    }

    return requestMessage;
  }

  Future<Response> makeRequest(
    Request request, {
    String url,
    int retries = 0,
    Duration timeLimit,
    Duration timeRetry = const Duration(milliseconds: 100),
    OnTimeout onTimeout,
    CancellationToken cancelToken,
    convert.Encoding encoding,
    Map<String, String> headers,
  }) async {
    if (await isConnected()) {
      Crypto crypto = Crypto(config.secretKey, config.publicKey);
      String requestMessage = _prepareRequestMessage(crypto, request);

      http.Response response = request.method == Method.POST
          ? await _post(url ?? config.url, requestMessage,
              timeLimit: timeLimit,
              onTimeout: onTimeout,
              cancelToken: cancelToken,
              encoding: encoding,
              headers: headers,
              retries: retries,
              timeRetry: timeRetry)
          : await _get(url ?? config.url, requestMessage,
              timeLimit: timeLimit,
              onTimeout: onTimeout,
              cancelToken: cancelToken,
              timeRetry: timeRetry,
              retries: retries,
              headers: headers);

      if (response != null) {
        if (response.statusCode == 200) {
          // OK
          var responseMessage = response.body;
          _print("raw response : " + responseMessage);

          if (config.encrypt) {
            responseMessage = crypto.decrypt(responseMessage);
            _print("decrypted response : " + responseMessage);
          }
          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print("json decoded : " + jsonResponse.toString());
            final r = Response(
              response.body,
              response.statusCode,
              result: Result(params: jsonResponse),
              state: ResponseState.SUCCESS,
            );
            return r;
          } catch (e) {
            _print("json decoded exception : " + e.toString());
            final r = Response(
              response.body,
              response.statusCode,
              result: Result(params: null),
              state: ResponseState.UNEXPECTED,
            );
            return r;
          }
        } else {
          // error
          _print(response.statusCode.toString() + " " + response.body);
          final r = Response(response.body, response.statusCode,
              result: null, state: ResponseState.ERROR);
          return r;
        }
      } else {
        _print("http response is null!");
        final r = Response("http response is null!", -1,
            result: null, state: ResponseState.ERROR);
        return r;
      }
    } else {
      return Response("Check your network connection!", -1,
          result: null, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  Future<Response> uploadFile(String fileKey, String filePath,
      {String url,
      Request request,
      StreamTransformer<List<int>, String> decoder}) async {
    if (await isConnected()) {
      http.MultipartRequest req = http.MultipartRequest(
          request != null ? request.method.value : 'POST',
          Uri.parse(url ?? config.url));
      req.files.add(await http.MultipartFile.fromPath(fileKey, filePath));

      Crypto crypto = Crypto(config.secretKey, config.publicKey);

      if (request != null) {
        String requestMessage = _prepareRequestMessage(crypto, request);

        req.fields["request"] = requestMessage;
        req.fields["os"] = _getOS();
      }

      var res = await req.send();
      if (res != null) {
        if (res.statusCode == 200) {
          var responseBody = await res.stream.bytesToString();
          var responseMessage = responseBody;
          _print("raw response : " + responseMessage);

          if (config.encrypt) {
            responseMessage = crypto.decrypt(responseMessage);
            _print("decrypted response : " + responseMessage);
          }
          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print("json decoded : " + jsonResponse.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: Result(params: jsonResponse),
              state: ResponseState.SUCCESS,
            );
          } catch (e) {
            _print("json decoded exception : " + e.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: Result(params: null),
              state: ResponseState.UNEXPECTED,
            );
          }
//        res.stream
//            .transform(decoder ?? convert.utf8.decoder)
//            .listen((responseBody) {
//          var responseMessage = responseBody;
//          _print("raw response : " + responseMessage);
//
//
//        });
        } else {
          // error
          _print(res.statusCode.toString() + " " + res.reasonPhrase);
          return Response(res.reasonPhrase, res.statusCode,
              result: null, state: ResponseState.ERROR);
        }
      } else {
        _print("http response is null!");
        return Response("http response is null!", -1,
            result: null, state: ResponseState.ERROR);
      }
    } else {
      return Response("Check your network connection!", -1,
          result: null, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  Future<Response> uploadFiles(
      {String url,
      Request request,
      StreamTransformer<List<int>, String> decoder,
      List<FileEntity> singleKey,
      MultipleFileEntity multiKey}) async {
    if (await isConnected()) {
      http.MultipartRequest req = http.MultipartRequest(
          request != null ? request.method.value : 'POST',
          Uri.parse(url ?? config.url));

      if (singleKey != null) {
        singleKey.forEach((file) async {
          req.files.add(await http.MultipartFile.fromPath(file.key, file.path));
        });
      }

      if (multiKey != null) {
        multiKey.paths.forEach((path) async {
          req.files.add(await http.MultipartFile.fromPath(multiKey.key, path));
        });
      }

      Crypto crypto = Crypto(config.secretKey, config.publicKey);
      if (request != null) {
        String requestMessage = _prepareRequestMessage(crypto, request);

        req.fields["request"] = requestMessage;
        req.fields["os"] = _getOS();
      }

      var res = await req.send();
      if (res != null) {
        if (res.statusCode == 200) {
          var responseBody = await res.stream.bytesToString();
          var responseMessage = responseBody;
          _print("raw response : " + responseMessage);

          if (config.encrypt) {
            responseMessage = crypto.decrypt(responseMessage);
            _print("decrypted response : " + responseMessage);
          }
          try {
            var jsonResponse = convert.jsonDecode(responseMessage);
            _print("json decoded : " + jsonResponse.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: Result(params: jsonResponse),
              state: ResponseState.SUCCESS,
            );
          } catch (e) {
            _print("json decoded exception : " + e.toString());
            return Response(
              responseBody,
              res.statusCode,
              result: Result(params: null),
              state: ResponseState.UNEXPECTED,
            );
          }
//        res.stream
//            .transform(decoder ?? convert.utf8.decoder)
//            .listen((responseBody) {
//
//        });
        } else {
          // error
          _print(res.statusCode.toString() + " " + res.reasonPhrase);
          return Response(res.reasonPhrase, res.statusCode,
              result: null, state: ResponseState.ERROR);
        }
      } else {
        _print("http response is null!");
        return Response("http response is null!", -1,
            result: null, state: ResponseState.ERROR);
      }
    } else {
      return Response("Check your network connection!", -1,
          result: null, state: ResponseState.NETWORK_CONNECTION);
    }
  }

  Future<http.Response> _post(String url, String requestMessage,
      {retries,
      timeRetry,
      timeLimit,
      onTimeout,
      cancelToken,
      headers,
      encoding}) async {
    var requestBody = {"request": requestMessage, "os": _getOS()};

    cancelToken?.throwIfCancellationRequested();
    return await RetryHelper.tryRun<http.Response>(() {
      return CancellationTokenSource.register(
          cancelToken,
          timeLimit == null
              ? client.post(url,
                  headers: headers,
                  body: requestBody,
                  encoding: encoding ?? convert.Encoding.getByName("utf-8"))
              : client
                  .post(url,
                      headers: headers,
                      body: requestBody,
                      encoding: encoding ?? convert.Encoding.getByName("utf-8"))
                  .timeout(timeLimit, onTimeout: onTimeout));
    }, cancelToken: cancelToken, timeRetry: timeRetry, retries: retries);
  }

  Future<http.Response> _get(String url, String requestMessage,
      {retries, timeRetry, timeLimit, onTimeout, cancelToken, headers}) async {
    cancelToken?.throwIfCancellationRequested();
    String queryUrl = url + "?request=" + requestMessage + "&os=" + _getOS();
    return await RetryHelper.tryRun<http.Response>(() {
      return CancellationTokenSource.register(
          cancelToken,
          timeLimit == null
              ? client.get(queryUrl, headers: headers)
              : client
                  .get(
                    queryUrl,
                    headers: headers,
                  )
                  .timeout(timeLimit, onTimeout: onTimeout));
    }, cancelToken: cancelToken, timeRetry: timeRetry, retries: retries);
  }

  String _getOS() {
    if (kIsWeb) return 'W';

    switch (defaultTargetPlatform) {
      case TargetPlatform.android:
        return "A";
      case TargetPlatform.iOS:
        return "I";
//      case TargetPlatform.windows:
//        return "D";
      default:
        return "Unkown";
    }
  }

  void _print(String message) {
    if (config.debugMode) print("APEX_FLUTTER_PLUGIN => " + message + "\r\n");
  }

  dispatch() {
    if (client != null) client.close();
  }

  Future<bool> isConnected() async {
    var connectivityResult = await (Connectivity().checkConnectivity());
    if (connectivityResult != ConnectivityResult.none || kIsWeb) {
      // I am connected to a mobile network.
      return true;
    }
    return false;
  }
}

class FileEntity {
  final String key;
  final String path;

  FileEntity(this.key, this.path);
}

class MultipleFileEntity {
  final String key;
  final List<String> paths;

  MultipleFileEntity(this.key, this.paths);
}

typedef OnTimeout = FutureOr<dynamic> Function();
